/**
 *  Provinding multisorting facilities for HTML tables.
 *
 *  Mickaël Wolff (http://lupusmic.org)
 *  22/06/2008
 *
 *  Licensing : AGPL <http://www.fsf.org/licensing/licenses/agpl-3.0.html>
 *
 *  Known issues :
 *    Not MSIE compliant because of event handling
 *    Don't work with Konqueror 3.5.9 due to a scope variable problem
 *      http://bugs.kde.org/show_bug.cgi?id=164804
 *
 *  Supposed issues :
 *    rowspan and colspan wrong behavior
 *    col and colgroup wrong behavior
 *
 *  Ideas :
 *    make possible to insert an ordered column between two others
 *    make interactive with other javascript
 *    add a class to specify non-data row (ie row captions)
 *    add a class to specify fixed-data row
 *    add an option to sort tbodies instead of rows
 *    add an option to prevent tbody swap informations
 *    display tbodies one by one, and provide methods to switch tbodies
 *    promote (or fall out) a row into a tbody on condition
 *    detect data type on value instead of /thead/tr/td[@class]
 *
 *  Tested on :
 *    Mozilla Firefox 3.0 (Linux)
 *    Opera 9.5 (Linux)
 *
 */

var multisort =
{ registry: Array()

// Deactivate tbody balancing by default
, MAX_ROW_BY_TBODY: -1
, MIN_ROW_BY_TBODY: -1

, FREEZE_CLASSNAME: 'freeze'

, manage: function(id)
	{
		multisort.registry[id] = new multisort.manager() ;
		multisort.registry[id].tableNode = document.getElementById(id) ;
		return multisort.registry[id] ;
	}
, callback_sort: function(ev)
	{
		/** \todo colspan support
		 *  \todo colgroup support
		 */
		var manager = multisort.registry[this.parentNode.parentNode.parentNode.getAttribute('id')] ;

		if(ev.ctrlKey)
			manager.unsort_by_hnode(this, ev) ;
		else
			manager.sort_by_hnode(this, ev) ;
	}
, manager: null
}

multisort.manager = function()
{
	// console.log(this) ;

	this.tableNode = null ;

	classOptions = {enable:true, name:'sorted', asc:'asc', desc:'desc'} ;
	rowsOptions = 
		{ tBody: {enable: false, min:multisort.MIN_ROW_BY_TBODY, max:multisort.MAX_ROW_BY_TBODY, mandatory:0}
		, freeze: {enable: false, className: multisort.FREEZE_CLASSNAME}
		} ;

	var columns = Array() ;
	var order = Array() ;

	/** This function compares two rows, to determine the
	 *  preceding of both.
	 *  It considers order sorting, and ascending
	 *
	 *  \param HTMLTableRowElement		first
	 *  \param HTMLTableRowElement		second
	 *  \return		 0 on equality
	 *  			-1 when first is before second
	 *  			 1 when first is after second
	 *
	 *  \bugs		Issue with some dates
	 */
	function compare_rows(first, second)
	{
		for(var i=0 ; i < order.length ; i++)
		{
			var cell_state = order[i] ;

			first_value = extract_typed_value(first.cells.item(cell_state.idx), cell_state.type) ;
			second_value = extract_typed_value(second.cells.item(cell_state.idx), cell_state.type) ;

			if(cell_state.asc)
			{
				if(first_value > second_value)
					return 1 ;
				else if(first_value < second_value)
					return -1 ;
			}
			else
			{
				if(first_value > second_value)
					return -1 ;
				else if(first_value < second_value)
					return 1 ;
			}
		}

		return 0 ;
	}

	/** \brief		Remove sorting on a column by head node
	 *  \param Node node
	 */
	this.unsort_by_hnode = function(node)
	{
		cell_order_idx = order.indexOf(columns[node.cellIndex]) ;
		var cell_state = order[cell_order_idx] ;
		order.splice(cell_order_idx, 1) ;

		if(classOptions.enable)
			unset_class_name.apply(this, [node, cell_state.asc]) ;

		for(var i=0 ; i<this.tableNode.tBodies.length ; i++)
			sort_rows.apply(this, [this.tableNode.tBodies.item(i).rows]) ;

		if(rowsOptions.tBody.enable)
			balance_bodies.apply(this) ;
	}

	/** \brief		Enable sorting on a column by head node
	 *  \param Node node
	 */
	this.sort_by_hnode = function(node)
	{
		if(this.tableNode.rows.length < 2)
			return ;

		var cell_state = columns[node.cellIndex] ;

		if(order.indexOf(cell_state) < 0)
			order.push(cell_state) ;

		cell_state.sorted = true ;
		cell_state.asc = cell_state.asc == false ? true : false ;

		if(classOptions.enable)
			set_class_name.apply(this, [node, cell_state.asc]) ;

		for(var i=0 ; i<this.tableNode.tBodies.length ; i++)
			sort_rows.apply(this, [this.tableNode.tBodies.item(i).rows]) ;

		if(rowsOptions.tBodies.enable)
			balance_bodies.apply(this) ;
	}

	/** \brief		Sort table rows
	 */
	function sort_rows(rowsCollection)
	{
		/*
		 * Bubble sort algorithm 
		 * <http://www.dailly.info/algorithmes-de-tri/bulle.php>
		 *
		 */
		do
		{
			inverted = false ;
			var length = rowsCollection.length ;
			for(var i=0 ; i < length - 1 ; i++)
			{

				var first = rowsCollection.item(i) ;
				var second = rowsCollection.item(i+1) ;

				// ignoring rows that must not be sort
				if(!is_sortable(first))
					continue ;
				if(!is_sortable(second))
				{
					// prevent an useless check for second as first
					i++ ;
					continue ;
				}

				// Todo make it body independent

				if(compare_rows(first, second) > 0)
				{
					first.parentNode.insertBefore(second, first) ;
					inverted = true ;
				}
			}
		} while(inverted) ;
	}

	function is_sortable(row)
	{
		return true ;
	}

	/** \brief		Assure that tbodies are well-balanced (ie: a 
	 *				maximum of rows by body)
	 *	\bugs		This must be used for inserting caption rows between
	 *				tbodies.
	 */
	var balance_bodies = function()
	{
		if(rowsOptions.tBody.mandatory < 0)
			return ;

		for(var i=0 ; i < this.tableNode.tBodies.length ; i++)
		{
			tb_current = this.tableNode.tBodies.item(i) ;
			tb_next = this.tableNode.tBodies.item(i+1) ;

			/* current tBody overflowed !
			 */
			while(tb_current.rows.length > rowsOptions.tBody.mandatory)
			{
				if(!tb_next)
					tb_next = this.tableNode.createTBody() ;

				if(tb_next.rows.length == 0)
					tb_next.appendChild(tb_current.rows.item(tb_current.rows.length - 1)) ;
				else
					tb_next.insertBefore(tb_current.rows.item(tb_current.rows.length - 1),
							tb_next.rows.item(0)) ;
			}

			/* current tBody underflowed !
			 */
			while(tb_current.rows.length < rowsOptions.tBody.mandatory && tb_next)
			{
				if(tb_next.rows.length == 0)
				{
					this.tableNode.removeChild(tb_next) ;
					tb_next = this.tableNode.tBodies.item(i+1) ;
				}

				if(tb_current.rows.length == 0)
					tb_current.appendChild(tb_next.rows.item(0)) ;
				else
					tb_current.insertBefore(tb_next.rows.item(0),
						tb_current.rows.item(tb_current.rows.length - 1)) ;
			}
		}
	}

	function set_class_name(node, asc)
	{
		if(node.className.indexOf(classOptions.name) > -1)
			node.className = asc
				? node.className.replace(classOptions.desc, classOptions.asc)
				: node.className.replace(classOptions.asc, classOptions.desc) ;
		else
			node.className += asc 
				? ' ' + classOptions.name + '-' + classOptions.asc
				: ' ' + classOptions.name + '-' + classOptions.desc ;
	}

	function unset_class_name(node, asc)
	{
		node.className = asc
			? node.className.replace(classOptions.name + '-' + classOptions.asc, '')
			: node.className.replace(classOptions.name + '-' + classOptions.desc, '') ;
	}

	/** I had to call it forceps :p
	 */
	function force_thead()
	{
		this.tableNode
			.createTHead()
			.appendChild(this.tableNode.rows.item(0)) ;

		return this.tableNode.tHead ;

		/* \todo
		var i = 0 ;

		var	rowSpanMax = row_span(this.tableNode.rows.item(i)) ;
		var	colSpanMax = col_span(this.tableNode.rows.item(i)) ;

		return this.tableNode.tHead ;
		*/
	}

	this.init = function()
	{
		if(!this.tableNode)
			return false ;

		/* If no thead element was provide, assume first row are heading
		 */
		if(!this.tableNode.tHead && !force_thead.apply(this)) return false ;

		if(rowsOptions.tBody.enable)
		{
			if(rowsOptions.tBody.mandatory < rowsOptions.tBody.min)
				rowsOptions.tBody.mandatory = rowsOptions.tBody.min ;
			else if(rowsOptions.tBody.mandatory > rowsOptions.tBody.max)
				rowsOptions.tBody.mandatory = rowsOptions.tBody.max ;
		}

		var head_cells = this.tableNode.tHead.rows.item(0).cells ;
		for(var i = 0 ; i < head_cells.length ; i++)
		{
			var cell = head_cells[i] ;
			var state = { idx: cell.cellIndex, type: null, sorted: false, asc: null} ;

			if(cell.className.indexOf('date') > -1)
				state.type = 'date' ;
			else if(cell.className.indexOf('numeric') > -1)
				state.type = 'numeric' ;
			else
				state.type = 'string' ;

			columns.push(state) ;
			
			// provide a bind option instead default event handler
			cell.addEventListener('click',		multisort.callback_sort,	false) ;
		}

		if(rowsOptions.tBody.enable)
			balance_bodies.apply(this) ;

		return this ;
	}
}

/** \brief  cast the node value in a javascript object, that modelize
 *			the data
 *
 *	\todo	make a plugeable object that can transtype everything we
 *			have to
 *	
 *	\param  node Node		the speciment
 *	\param  type String		a case insensitive identifier that can take
 *							those values :
 *								- numeric or number
 *								- date
 *								- string (as default)
 *	\return					the fine object
 */
function extract_typed_value(node, type)
{
	type = type.toLowerCase() ;

	if(type == 'date')
		value = l10n_html_date(node) ;
	else if(type == 'numeric' || type == 'number')
		value = new Number(node.textContent) ;
	else
		value = node.textContent ;

	return value ;
}

/** \todo find a finest name
 *	\todo make a class that inherit from Date
 *	\todo support more l10n 
 */
var l10n_html_date = function(node)
{
	var pieces = node.textContent.split('/') ;
	var lang = node.lang ;

	var p_node = node.parentNode ;
	while(p_node && lang.length==0)
	{
		lang = p_node.lang ;
		p_node = p_node.parentNode ;
	}

	var day ;

	if(lang == 'fr')
		day = new Date(pieces[2], pieces[1], pieces[0]) ;
	else if(lang == 'jp')
		day = new Date(pieces[0], pieces[1], pieces[2]) ;
	else
	{
		lang = 'en' ;
		day = new Date(pieces[2], pieces[0], pieces[1]) ;
	}

	return day ;
}

